/** 
 * Pääohjelma.
 * Tarvitsee toimiakseen niiden moduulien otsikkotiedostot, 
 * joiden aliohjelmia tai dataa tarvitaan suoraan. Konekielinen
 * toteutus tarvitaan vasta linkitysvaiheessa.
 */
#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <math.h>

#include "koktaulu.h"
#include "tietue.h"
#include "merkkijono.h"

/* Tässä asetetaan C:mäisesti vakioita. Moduuleissa toteutettujen
 * "yleiskäyttöisten" aliohjelmien ei pitäisi joutua tietämään
 * tällaisista, vaan niille nämä on ilmoitettava parametrina
 * kutsuissa.
 */

#define TAULUN_KOKO 5
#define MAX_JONON_KOKO 100
#define MAX_PIKKUJONON_KOKO 10


/** Testailee koktaulu -moduulia. Palauttaa virheellisesti toimivien
 *  testikeissien määrän. Tulostelee havainnollistavia juttuja
 *  lisäksi.
 */
int taulukkokokkeilut(){
    int virheiden_lkm = 0;
    int64_t a[TAULUN_KOKO];
    int64_t tA[3]   = {-1,   5,   9};  // Testitaulukko A
    int64_t tB[3]   = {14, -15, -99};  // Testitaulukko B
    int64_t tApB[3] = {13, -10, -90};  // Oletetut summat A+B

    /* HUOMAA mitä tapahtuu, kun ajat ohjelmaa useita kertoja
     * peräkkäin!  C:ssä tilan varaaminen muuttujille ei alusta arvoja
     * nollaksi tai yhtään miksikään. Arvo on se, joka
     * muistiosoitteessa sattui olemaan! Ohjelmoijalla on kontrolli ja
     * toisaalta vastuu kaikesta. Muista alustaa muuttujat C:ssä
     * tarkoituksenmukaisesti ennen käyttöä. (Esim. Javassahan
     * luvataan, että muuttuja on nolla tai viitemuuttuja 'null' aina
     * aluksi)
     */
    printf("Taulun sisältö aluksi: ");
    koktaulu_tulosta(a, TAULUN_KOKO, stdout);
    printf("Summa aluksi = %" PRId64 "\n", koktaulu_summa(a, TAULUN_KOKO));
    koktaulu_tayta(a, TAULUN_KOKO, 9);
    printf("Summa lopuksi = %" PRId64 "\n", koktaulu_summa(a, TAULUN_KOKO));
    printf("Taulun sisältö lopuksi: ");
    koktaulu_tulosta(a, TAULUN_KOKO, stdout);
    printf("Testitaulu A: ");
    koktaulu_tulosta(tA, 3, stdout);
    printf("Testitaulu B: ");
    koktaulu_tulosta(tB, 3, stdout);

    /** Huomaa, että muistiosoitteena vietyyn täkäläiseen taulukon
        tulee tässä muutos (jos olet osannut sen koodata
        täydennystehtävänä)*/
    koktaulu_summa_alkioittain_bhen(tA,tB,3);

    printf("Laskemasi alkioittainen summa 'A+B': ");
    koktaulu_tulosta(tB, 3, stdout);
    printf("Testin edellyttämä alkioittainen summa: ");
    koktaulu_tulosta(tApB, 3, stdout);
    if (!koktaulu_samansisaltoisia(tB,tApB,3)){
        printf("Laskettu ja edellytetty eivät ole samoja!\n");
        virheiden_lkm++;
    }
    return virheiden_lkm;
}



/** Testailee tietue -moduulia. Palauttaa virheellisesti toimivien
 *  testikeissien määrän. Tulostelee havainnollistavia juttuja
 *  lisäksi.
 */
int tietuekokkeilut(){
    int virheiden_lkm = 0;
    vect3d_t piste = {1.0, 2.0, 3.4};
    vect3d_t a = {-1.5, 200.0, -30.0};
    vect3d_t b = {2.0, -100.0, 2.0};
    /*(-1.5*2) + (-100*200) + (-30*2) = -20063 */
    double odotettu_pistetulo = -20063;

    /*Huomaa, että nyt on varattu tietueelle tilaa pinosta. Aliohjelmaan
     * vaaditaan osoitin, joten käytetään "reference operatoria" eli
     * viitteen ottamista eli pointterin tekemistä muuttujaan. Se on tuo
     * syntaksi &piste tuolla. Siis et-merkillä merkitty operaattori &
     * ottaa oikealla puolellaan olevan muuttujan muistiosoitteen, jonka
     * voi sijoittaa sopivaan osoitinmuuttujaan:
     */
    printf("Pisteen 'piste' koordinaatit (%.4f %.4f %.4f), etäisyys origosta %.4f\n",
           piste.x, piste.y, piste.z, vect3d_norm(&piste));
    printf("Vec1 = (%.4f %.4f %.4f), Vec2 = (%.4f %.4f %.4f)\n",
           a.x, a.y, a.z, b.x, b.y, b.z);

    double dot_ab = vect3d_pistetulo(&a, &b);
    printf("Laskemasi pistetulo 'dot(Vec1,Vec2)' = %.4f)\n", dot_ab);
    printf("Odotettu tulos oli %.4f\n", odotettu_pistetulo);
    if (!(fabs(dot_ab - odotettu_pistetulo) < .000001)){
        printf("Ei täsmää!\n");
        virheiden_lkm++;
    }
    return virheiden_lkm;
}


/** Testailee merkkijono -moduulia. Palauttaa virheellisesti toimivien
 *  testikeissien määrän. Tulostelee havainnollistavia juttuja
 *  lisäksi.
 */
int merkkijonokokkeilut(){
    int virheiden_lkm = 0;
    int paluuarvo;
    char munjono[MAX_JONON_KOKO] = "Zippadai dippa daabiduu";
    char odotettu_jono[MAX_JONON_KOKO] = "ZaAdAa daA dAAbad";

    char testijono2[MAX_PIKKUJONON_KOKO] = "avvaava";
    char testijono3[MAX_PIKKUJONON_KOKO+1] = "aaaaaaaaaa";
    
    printf("Jono aluksi (pituus %" PRId64 "merkkiä) = %s\n", 
           (int64_t) strlen(munjono), munjono);

    /* Huomaa seuraavassa miten osoitinparametri toimii: Osoitetun
     * muistin tila voi muuttua (aliohjelmalla on
     * "sivuvaikutus"). Tällä tavoin ohjelmointi tietorakenteilla ja
     * aliohjelmilla on melkein sama asia kuin olioiden
     * käyttö. Ajattele merkkitaulua "oliona" ja aliohjelmia
     * mjono_jotakin() "metodeina" joilla käsitellään mjono-"luokan"
     * "instansseja". "Instanssin viite" annetaan ensimmäisenä
     * parametrina "oliota" käsittelevän "metodin" toteutukselle:
     */
    mjono_korvaa_merkki(munjono, MAX_JONON_KOKO, 'a', 'A');
    mjono_korvaa_merkki(munjono, MAX_JONON_KOKO, 'i', 'a');
    printf("Jono sittemmin = %s\n", munjono);

    /* Paljonkaan eroa ei ole oliokieliseen ilmaukseen
     * munjono.korvaa_merkki('a','A'). Näin ajateltuna
     * (toteutustasolla) olio-ohjelmointi on erityistapaus
     * imperatiivisesta tietorakenneohjelmoinnista.  Mutta
     * oliomallinnus suunnitteluperiaatteena ja ns. "kaikkine
     * herkkuineen" (rajapinnat, perintä, polymorfismi, ...)
     * kielityökalun kautta tuettuna on tietysti laajempi ja hieman
     * syvällisemmin eri asia.
     */

    mjono_poista_merkki(munjono, MAX_JONON_KOKO, 'p');
    mjono_poista_merkki(munjono, MAX_JONON_KOKO, 'u');
    printf("Jono poistojen jälkeen = '%s'\n", munjono);
    printf("Odotettu lopputulos    = '%s'\n", odotettu_jono);

    if (strcmp(munjono,odotettu_jono) != 0){
        printf("Muunnettu ja odotettu jono eivät täsmää\n");
        virheiden_lkm++;
    }

    /* Poistellaan loputkin ja katotaan, ettei jää mitään: */
    mjono_poista_merkki(munjono, MAX_JONON_KOKO, 'Z');
    mjono_poista_merkki(munjono, MAX_JONON_KOKO, 'a');
    mjono_poista_merkki(munjono, MAX_JONON_KOKO, ' ');
    mjono_poista_merkki(munjono, MAX_JONON_KOKO, 'd');
    mjono_poista_merkki(munjono, MAX_JONON_KOKO, 'b');
    mjono_poista_merkki(munjono, MAX_JONON_KOKO, 'A');
    /* Ja tuo ei tietysti saisi tehdä mitään: */
    mjono_poista_merkki(munjono, MAX_JONON_KOKO, '\0');
    
    if (strlen(munjono) > 0){
        printf("Kaikkien poistojen jälkeen piti olla tyhjä jono. Nyt oli vielä '%s'\n",munjono);
        virheiden_lkm++;
    }

    /* Tarkastetaan erityistapausten paluuarvot: */
    paluuarvo = mjono_poista_merkki(testijono2, MAX_PIKKUJONON_KOKO, 'a');
    if (paluuarvo != 3) {
        printf("Kun jonosta 'avvaava' poistetaan 'a', uuden jonon 'vvv' pituus olisi 3\n"
               "Nyt kuitenkin vaikuttaa palautuvan %d!\n", paluuarvo);
        virheiden_lkm++;
    }

    paluuarvo = mjono_poista_merkki(testijono2, MAX_PIKKUJONON_KOKO, '\0');
    if (paluuarvo != -1){
        printf("Laittomalla parametrilla paluuarvon tulisi olla -1\n"
               "Nyt kuitenkin vaikuttaa palautuvan %d\n", paluuarvo);
        virheiden_lkm++;
    }

    paluuarvo = mjono_poista_merkki(testijono3, MAX_PIKKUJONON_KOKO, 'a');
    if (paluuarvo != -2){
        printf("Laittomalla merkkijonolla paluuarvon tulisi olla -2\n"
               "Nyt kuitenkin vaikuttaa palautuvan %d\n", paluuarvo);
        virheiden_lkm++;
            
    }

    
    return virheiden_lkm;
}


/** Testailee moduuleita; palauttaa selvästi virheellisesti toimivien
 *  koodien lukumäärän.
 */
int onnistuuko_kokkeilut(){
    int virh_lukum = 0;
    virh_lukum += taulukkokokkeilut();
    printf("\n"); /*Ehkä selkeämpi tuloste, jos on tyhjää välissä?*/
    virh_lukum += tietuekokkeilut();
    printf("\n");
    virh_lukum += merkkijonokokkeilut();
    printf("\n");

    return virh_lukum;
}

/** C-kielen sovittu "entry point" eli main()-aliohjelma. */
int main(int argc, char *argv[]){
    printf("--------------------------------------------------------\n"
           "Hei. Tämä ohjelma sai %d %s.\n"
           "Ohjelman sijaintiin liittyvä merkkijono eli argv[0] oli '%s'\n",
           argc, argc==1?"argumentin":"argumenttia", argv[0]);
    printf("--------------------------------------------------------\n");
    printf("Kokeillaan \"moduuleihin\" kirjoitettuja koodeja: \n");
           
    int nvirheita = onnistuuko_kokkeilut();

    if (nvirheita > 0){
        fprintf(stderr, "EI VIELÄ KUNNOSSA: %d kpl\n",nvirheita);
    } else {
        fprintf(stdout, "Toimii niinkuin haluttiinkin\n");
    }

    return nvirheita;
}
